#pragma once

#include <cppcore/conversion/enum.h>
#include <cppcore/conversion/convert_cast.h>

#include "exception.h"

namespace cpprtti
{

    struct error_code_enum_values
    {
        using enum_type              = error_code;
        using enum_value_pair_type   = std::pair<enum_type, std::string>;
        using enum_value_vector_type = std::vector<enum_value_pair_type>;

        static const enum_value_vector_type& get_enum_values()
        {
            static const enum_value_vector_type values({
                { error_code::unknown,                          "Unknown error" },

                { error_code::none,                             "No error" },

                /* registry */
                { error_code::registry_error,                   "Registry error" },
                { error_code::registry_type_not_found,          "Unable to find the requested type!" },
                { error_code::registry_id_already_exsists,      "Unable to create type: The type ID does already exist!" },
                { error_code::registry_name_already_exsists,    "Unable to create type: The type name does already exist!" },
                { error_code::registry_type_does_not_match,     "Unable to get type: The implementation of the type does not match!" },

                /* variant */
                { error_code::variant_error,                    "Variant error" },
                { error_code::variant_is_empty,                 "The variant does not store any value!" },
                { error_code::variant_type_mismatch,            "The variant does not store the requested type!" },
                { error_code::variant_invalid_cast,             "The value stored in the variant could not be cast to the requested type!" },

                /* class */
                { error_code::class_error,                      "Class error" },
                { error_code::class_member_already_exists,      "Unable to create member: Name is already in use!" },
                { error_code::class_type_mismatch,              "Type does not match the." },
                { error_code::class_member_not_readable,        "Member variable is not readable!" },
                { error_code::class_member_not_writable,        "Member variable is not writable!" },
            });
            return values;
        }
    };

    using error_code_enum_conversion_traits = ::cppcore::enum_conversion_traits<error_code, error_code_enum_values>;
    using error_code_enum_conversion        = ::cppcore::enum_conversion       <error_code, error_code_enum_conversion_traits>;

    std::string get_error_code_str(error_code err)
        { return error_code_enum_conversion::to_string(err); }

    /* exception */

    exception::exception(
        error_code          p_error)
        : cppcore::exception()
        , error             (p_error)
        , detail            ()
        { }

    exception::exception(
        error_code          p_error,
        const std::string&  p_detail)
        : cppcore::exception(p_detail)
        , error             (p_error)
        , detail            (p_detail)
        { }

    void exception
        ::print_message(std::ostream& os) const
    {
        os  << get_error_code_str(error)
            << '('
            << cppcore::convert_cast<std::underlying_type_t<error_code>>(error)
            << "): "
            << message;
    }

}
