#pragma once

#include <memory>
#include <vector>

#include "helper.h"

namespace cpprtti
{

    struct storage_container
    {
    private:
        using item_ptr_s  = std::shared_ptr<void>;
        using item_vector = std::vector<item_ptr_s>;

    private:
        item_vector _items;

    public:
        /**
         * @brief Get the requested storage or nullptr if the storage does not exist.
         */
        template<typename T>
        T* storage();

        /**
         * @brief Get the requested storage or nullptr if the storage does not exist.
         */
        template<typename T>
        T* storage(bool can_create);

        /**
         * @brief Create a new storage.
         */
        template<typename T>
        T* storage(const std::shared_ptr<T>& item);

        /**
         * @brief Get the requested storage or nullptr if the storage does not exist.
         */
        template<typename T>
        const T* storage() const;
    };

}
