#pragma once

#include <cppmp/core/modifier.h>

#include "storage.h"

namespace cpprtti
{

    /* storage_container */

    template<typename T>
    T* storage_container::storage()
    {
        using storage_type = cppmp::decay_t<T>;
        auto id = get_storage_id<storage_type>();
        return (id < _items.size())
            ? static_cast<storage_type*>(_items[id].get())
            : nullptr;
    }

    template<typename T>
    T* storage_container::storage(bool can_create)
    {
        using storage_type = cppmp::decay_t<T>;
        auto id = get_storage_id<storage_type>();
        if (    can_create
            && (    id >= _items.size()
                || !_items[id]))
        {
            _items.resize(id + 1);
            _items[id] = std::make_shared<storage_type>();
        }
        return (id < _items.size())
            ? static_cast<storage_type*>(_items[id].get())
            : nullptr;
    }

    template<typename T>
    T* storage_container::storage(const std::shared_ptr<T>& item)
    {
        using storage_type = cppmp::decay_t<T>;
        auto  id = get_storage_id<storage_type>();
        _items.resize(id + 1);
        _items[id] = item;
        return (id < _items.size())
            ? static_cast<storage_type*>(_items[id].get())
            : nullptr;
    }

    template<typename T>
    const T* storage_container::storage() const
    {
        using storage_type = cppmp::decay_t<T>;
        auto id = get_storage_id<storage_type>();
        return (id < _items.size())
            ? static_cast<storage_type*>(_items[id].get())
            : nullptr;
    }

}
