#pragma once

#include <map>
#include <vector>
#include <memory>

#include "type.h"
#include "member/member.h"

namespace cpprtti
{

    struct class_type
        : public type
    {
    private:
        using member_ptr_u      = std::unique_ptr<member>;
        using member_map        = std::map<std::string, member_ptr_u>;
        using class_type_vector = std::vector<const class_type*>;

    private:
        member_map          _members;
        class_type_vector   _base_classes;

    public:
        /**
         * @brief Constructor.
         */
        inline class_type(
            cpprtti::registry&  p_registry,
            type_id             p_id,
            const std::string&  p_name);

        /**
         * @brief Get the vector of base classes.
         */
        inline const class_type_vector& base_classes() const;

        /**
         * @brief Get a map of all registered members.
         */
        inline const member& member(const std::string& p_name) const;

        /**
         * @brief Get a map of all registered members.
         */
        inline const member_map& members() const;

        /**
         * @brief Register a new member variable.
         */
        template<typename... T_args>
        inline auto& add_member_variable(
            const std::string&  p_name,
            T_args&&...         p_args);

        /**
         * @brief Register a new member method.
         */
        template<typename... T_args>
        inline auto& add_member_method(
            const std::string&  p_name,
            T_args&&...         p_args);

        /**
         * @brief Add a base class to this class type.
         */
        inline auto& inherits(
            const class_type&   p_base_class);
    };

}
