#pragma once

#include "class_type.h"
#include "member/member_method.h"
#include "member/member_variable.h"

namespace cpprtti
{

    /* class_type */

    class_type::class_type(
        cpprtti::registry&  p_registry,
        type_id             p_id,
        const std::string&  p_name)
        : type(p_registry, p_id, p_name, rtti_type_t::class_type)
        { }

    const class_type::class_type_vector& class_type::base_classes() const
        { return _base_classes; }

    const member& class_type::member(
        const std::string& p_name) const
    {
        auto it = _members.find(p_name);
        if (it == _members.end())
        {
            throw exception(
                error_code::class_member_does_not_exist,
                cppcore::string_builder()
                    << "Member with the name '"
                    << p_name
                    << "' does not exists");
        }
        return *it->second;
    }

    const class_type::member_map& class_type
        ::members() const
        { return _members; }

    template<typename... T_args>
    auto& class_type
        ::add_member_variable(
            const std::string&  p_name,
            T_args&&...         p_args)
    {
        auto it = _members.find(p_name);
        if (it != _members.end())
        {
            throw exception(
                error_code::class_member_already_exists,
                cppcore::string_builder()
                    << "Member with the name '"
                    << p_name
                    << "' already exists");
        }

        auto  ptr = make_member_variable(*this, p_name, std::forward<T_args>(p_args)...);
        auto& ret = *ptr;
        _members.emplace_hint(it, p_name, std::move(ptr));

        return ret;
    }

    template<typename... T_args>
    auto& class_type
        ::add_member_method(
            const std::string&  p_name,
            T_args&&...         p_args)
    {
        auto it = _members.find(p_name);
        if (it != _members.end())
        {
            throw exception(
                error_code::class_member_already_exists,
                cppcore::string_builder()
                    << "Member with the name '"
                    << p_name
                    << "' already exists");
        }

        auto ptr = make_member_method(*this, p_name, std::forward<T_args>(p_args)...);
        auto& ret = *ptr;
        _members.emplace_hint(it, p_name, std::move(ptr));

        return ret;
    }

    auto& class_type
        ::inherits(
            const class_type& p_base_class)
    {
        _base_classes.emplace_back(&p_base_class);
        return *this;
    }

}
