#pragma once

#include <string>

#include "../../types/type.h"
#include "../../misc/storage.h"

namespace cpprtti
{

    enum class member_type
    {
        unknown = 0,
        variable,
        method,
    };

    struct member
        : public storage_container
    {
    protected:
        cpprtti::type&  _owner;
        std::string     _name;
        member_type     _member_type;

    public:
        /**
         * @brief Constructor.
         */
        inline member(
            cpprtti::type&      p_owner,
            const std::string&  p_name,
            member_type         p_member_type);

        /**
         * @brief Destructor.
         */
        virtual ~member() = default;

        /**
         * @brief Type that owns this member.
         */
        inline const cpprtti::type& owner() const;

        /**
         * @brief Get the name of the member.
         */
        inline const std::string& name() const;

        /**
         * @brief Type of the member.
         */
        inline cpprtti::member_type member_type() const;
    };

}
