#pragma once

#include <cppmp/traits/lambda_traits.h>

#include "member_method.h"
#include "../../misc/variant.h"

namespace cpprtti
{

    /* member_method */

    member_method::member_method(
            cpprtti::type&      p_owner,
            const std::string&  p_name)
        : member(
            p_owner,
            p_name,
            cpprtti::member_type::method)
        { }

    template<typename T_object, typename... T_args>
    variant member_method
        ::invoke(T_object&& obj, T_args&&... args) const
    {
        return invoke_impl(
            make_variant(_owner.registry(), std::forward<T_object>(obj)),
            std::vector<variant>({
                make_variant(_owner.registry(), std::forward<T_args>(args))...
            }));
    }

    namespace __impl
    {

        /* member_method_builder - default */

        template<typename X, typename>
        struct member_method_builder
        {
            using is_default = cppmp::true_t;

            template<typename... T_args>
            static constexpr decltype(auto) apply(T_args&&...)
                { static_assert(sizeof...(T_args) == -1, "Invalid parameters for make_member_method(...)!"); }
        };

        /* member_method_builder - member method */

        template<typename T_method>
        struct member_method_builder<
            cppmp::list<class_type&, const std::string&, T_method>,
            cppmp::enable_if_t<
                    cppmp::is_valid_v<cppmp::lambda_traits<T_method>>
                && !cppmp::lambda_traits<T_method>::is_static_v
            >
        >
        {
            using method_type  = T_method;
            using lambda_trait = cppmp::lambda_traits<method_type>;
            using object_type  = typename lambda_trait::object_type;

            struct member_impl
                : public member_method
            {
            private:
                method_type _member;

            public:
                inline member_impl(
                    cpprtti::type&      p_owner,
                    const std::string&  p_name,
                    method_type         p_member)
                    : member_method(p_owner, p_name)
                    , _member(p_member)
                {
                    if (_owner.id() != get_type_id<cppmp::decay_t<object_type>>())
                    {
                        throw exception(
                            error_code::class_type_mismatch,
                            "Object type of the member method does not match the type of the owner!");
                    }
                }

            protected:
                variant invoke_impl(const variant& obj, const std::vector<variant>& args) const override
                    { return invoke_helper(obj, args, std::make_index_sequence<lambda_trait::argument_count_v> { }); }

            private:
                template<size_t... I>
                inline variant invoke_helper(const variant& obj, const std::vector<variant>& args, std::index_sequence<I...>) const
                {
                    return make_variant(
                        _owner.registry(),
                        (obj.get<object_type&>().*_member)(
                            args[I].get<typename lambda_trait::template argument_t<I>>()...
                        )
                    );
                }
            };

            template<typename X_member>
            static constexpr decltype(auto) apply(type& owner, const std::string& name, X_member&& member)
            {
                return std::make_unique<member_impl>(
                    owner,
                    name,
                    std::forward<X_member>(member));
            }
        };

    }

}
