#pragma once

#include <cppmp/misc/generic_predicate.h>

#include "member.h"
#include "../../misc/variant.h"

namespace cpprtti
{

    namespace __impl
    {

        template<typename X, typename = void>
        struct member_variable_builder;

    }

    struct member_variable
        : public member
    {
    protected:
        cpprtti::type&  _type;
        bool            _is_readale    { true };
        bool            _is_writable   { true };

    public:
        /**
         * @brief Constructor.
         */
        inline member_variable(
            cpprtti::type&      p_owner,
            const std::string&  p_name,
            cpprtti::type&      p_type);

        /**
         * @brief RTTI type of that member represents.
         */
        inline const cpprtti::type& type() const;

        /**
         * @brief Check if the variable is readable.
         */
        inline bool is_readale() const;

        /**
         * @brief Check if the variable is writable.
         */
        inline bool is_writable() const;

        /**
         * @brief Get the current value of the member.
         */
        template<typename T_object>
        inline variant get(T_object& obj) const;

        /**
         * @brief Get the current value of the member.
         */
        inline variant get(const variant& obj) const;

        /**
         * @brief Set the new value of the member.
         */
        template<typename T_object, typename T_value>
        inline void set(T_object& obj, const T_value& value) const;

        /**
         * @brief Set the new value of the member.
         */
        inline void set(const variant& obj, const variant& value) const;

    protected:
        /**
         * @brief Get the current value of the member.
         */
        virtual variant get_impl(const variant& obj) const = 0;

        /**
         * @brief Set the new value of the member.
         */
        virtual void set_impl(const variant& obj, const variant& value) const = 0;
    };

    constexpr decltype(auto) make_member_variable = cppmp::generic_predicate<__impl::member_variable_builder> { };

}
