#pragma once

#include "type.h"
#include "../registry.h"

namespace cpprtti
{

    /* type */

    type::type(
        cpprtti::registry&  p_registry,
        size_t              p_id,
        const std::string&  p_name,
        rtti_type_t         p_rtti_type)
        : _registry (p_registry)
        , _id       (p_id)
        , _name     (p_name)
        , _rtti_type(p_rtti_type)
        { }

    cpprtti::registry& type::registry()
        { return _registry; }

    const cpprtti::registry& type::registry() const
        { return _registry; }

    type_id type::id() const
        { return _id; }

    const std::string& type::name() const
        { return _name; }

    rtti_type_t type::rtti_type() const
        { return _rtti_type; }

    template<typename T_to>
    const type::converter* type::convert_to() const
    {
        using to_type = cppmp::decay_t<T_to>;
        return convert_to(get_type_id<to_type>());
    }

    const type::converter* type::convert_to(type_id id) const
    {
        auto it = _converters.find(id);
        return it == _converters.end()
            ? nullptr
            : &it->second;
    }

    template<typename T_from>
    const type::converter* type::convert_from() const
    {
        using from_type = cppmp::decay_t<T_from>;
        return convert_from(get_type_id<from_type>());
    }

    const type::converter* type::convert_from(type_id id) const
    {
        auto* type = _registry.find(id);
        if (!type)
            return nullptr;

        auto it = type->_converters.find(this->_id);
        return it == type->_converters.end()
            ? nullptr
            : &it->second;
    }

}
