#include <gtest/gtest.h>

#include <cpprtti.h>

using namespace cpprtti;

TEST(variant_test, empty)
{
    variant v;
    EXPECT_TRUE (v.empty());
    EXPECT_FALSE(static_cast<bool>(v));
}

TEST(variant_test, store_normal_value)
{
    registry reg;
    int i = 6;
    auto v = make_variant(reg, static_cast<int>(i));

    EXPECT_FALSE    (v.is_reference());
    EXPECT_FALSE    (v.is_const());

    EXPECT_EQ       (v.get<int>(), 6);
    EXPECT_EQ       (v.get<int&>(), 6);
    EXPECT_EQ       (v.get<int&&>(), 6);

    EXPECT_EQ       (v.get<const int>(), 6);
    EXPECT_EQ       (v.get<const int&>(), 6);
    EXPECT_EQ       (v.get<const int&&>(), 6);
}

TEST(variant_test, store_constant_value)
{
    registry reg;
    int i = 6;
    auto v = make_variant(reg, static_cast<int>(i));

    EXPECT_FALSE    (v.is_reference());
    EXPECT_FALSE    (v.is_const());

    EXPECT_EQ       (v.get<int>(), 6);
    EXPECT_EQ       (v.get<int&>(), 6);
    EXPECT_EQ       (v.get<int&&>(), 6);

    EXPECT_EQ       (v.get<const int>(), 6);
    EXPECT_EQ       (v.get<const int&>(), 6);
    EXPECT_EQ       (v.get<const int&&>(), 6);
}

TEST(variant_test, store_lvalue_reference)
{
    registry reg;
    int i = 6;
    auto v = make_variant(reg, static_cast<int&>(i));

    EXPECT_TRUE     (v.is_reference());
    EXPECT_FALSE    (v.is_const());

    EXPECT_EQ       (v.get<int>(), 6);
    EXPECT_EQ       (v.get<int&>(), 6);
    EXPECT_EQ       (v.get<int&&>(), 6);

    EXPECT_EQ       (v.get<const int>(), 6);
    EXPECT_EQ       (v.get<const int&>(), 6);
    EXPECT_EQ       (v.get<const int&&>(), 6);
}

TEST(variant_test, store_constant_lvalue_reference)
{
    registry reg;
    int i = 6;
    auto v = make_variant(reg, static_cast<const int&>(i));

    EXPECT_TRUE     (v.is_reference());
    EXPECT_TRUE     (v.is_const());

    EXPECT_EQ       (v.get<int>(), 6);
    EXPECT_ANY_THROW(v.get<int&>());
    EXPECT_ANY_THROW(v.get<int&&>());

    EXPECT_EQ       (v.get<const int>(), 6);
    EXPECT_EQ       (v.get<const int&>(), 6);
    EXPECT_ANY_THROW(v.get<const int&&>());
}

TEST(variant_test, store_rvalue_reference)
{
    registry reg;
    int i = 6;
    auto v = make_variant(reg, std::move(i));

    EXPECT_FALSE    (v.is_reference());
    EXPECT_FALSE    (v.is_const());

    EXPECT_EQ       (v.get<int>(), 6);
    EXPECT_EQ       (v.get<int&>(), 6);
    EXPECT_EQ       (v.get<int&&>(), 6);

    EXPECT_EQ       (v.get<const int>(), 6);
    EXPECT_EQ       (v.get<const int&>(), 6);
    EXPECT_EQ       (v.get<const int&&>(), 6);
}
