#include <gtest/gtest.h>

#include <cpprtti.h>

using namespace cpprtti;

struct test_class
{
public:
    int         i;
    std::string read_only;
    std::string write_only;
    std::string read_write;

public:
    inline const std::string& get_read_only() const
        { return read_only; }

    inline void set_write_only(const std::string& value)
        { write_only = value; }

    inline const std::string& get_read_write() const
        { return read_write; }

    inline void set_read_write(const std::string& value)
        { read_write = value; }

    inline int add(int value)
        { return i += value; }
};

TEST(class_type_tests, member_variable_read_write_getter)
{
    registry reg;
    auto& c = reg.get<test_class>();

    test_class obj;
    obj.i = 10;

    auto& m = c.add_member_variable("i", &test_class::i);
    auto  v = m.get(obj);
    EXPECT_TRUE (m.is_readale());
    EXPECT_TRUE (m.is_writable());
    EXPECT_TRUE (v.is_reference());
    EXPECT_FALSE(v.is_const());
    EXPECT_EQ   (v.get<int>(), 10);
}

TEST(class_type_tests, member_variable_read_only_getter)
{
    registry reg;
    auto& c = reg.get<test_class>();

    test_class obj;
    obj.read_only = "read_only";

    auto& m = c.add_member_variable("read_only", &test_class::get_read_only);
    auto  v = m.get(obj);
    EXPECT_TRUE (m.is_readale());
    EXPECT_FALSE(m.is_writable());
    EXPECT_TRUE (v.is_reference());
    EXPECT_TRUE (v.is_const());
    EXPECT_EQ   (v.get<std::string>(), "read_only");
}

TEST(class_type_tests, member_variable_write_only_setter)
{
    registry reg;
    auto& c = reg.get<test_class>();

    test_class obj;

    auto& m = c.add_member_variable("write_only", &test_class::set_write_only);
    EXPECT_FALSE(m.is_readale());
    EXPECT_TRUE (m.is_writable());
    m.set(obj, std::string("fuuu"));
    EXPECT_EQ(obj.write_only, "fuuu");
}

TEST(class_type_tests, member_variable_read_write_getter_setter)
{
    registry reg;
    auto& c = reg.get<test_class>();

    test_class obj;
    obj.read_write = "read_write";

    auto& m = c.add_member_variable("read_write", &test_class::get_read_write, &test_class::set_read_write);
    auto  v = m.get(obj);
    EXPECT_TRUE (m.is_readale());
    EXPECT_TRUE (m.is_writable());
    EXPECT_TRUE (v.is_reference());
    EXPECT_TRUE (v.is_const());
    EXPECT_EQ   (v.get<std::string>(), "read_write");
    m.set(obj, std::string("fuuu"));
    EXPECT_EQ(obj.read_write, "fuuu");
}

TEST(class_type_tests, member_method)
{
    registry reg;
    auto& c = reg.get<test_class>();

    test_class obj;
    obj.i = 10;

    auto& m = c.add_member_method("add", &test_class::add);
    auto  v = m.invoke(obj, 5);

    EXPECT_EQ(v.get<int>(), 15);
    EXPECT_EQ(obj.i,        15);
}
